(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  Guillaume Allais (2011)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2011-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory RPO_Impl
imports 
  RPO_More
  Efficient_Weighted_Path_Order.RPO_Mem_Impl
begin

section \<open>LPO Closures\<close>

context rpo_with_assms
begin

context
  fixes F :: "('f \<times> nat) set"
  assumes pr_gtotal: "\<And>f g. f \<in> F \<Longrightarrow> g \<in> F \<Longrightarrow> f = g \<or> fst (prc f g) \<or> fst (prc g f)"
  and lpo: "\<And> f. f \<in> F \<Longrightarrow> c f = Lex"
  and arity_restr:"\<forall> (f :: 'f) i. (f, i) \<in> funas_term s \<union> funas_term t \<longrightarrow> i \<le> n"
begin

text \<open>The next definition forms a closure of a given relation gt with respect to a LPO
      (following Martin/Nipkow CADE 1990, where gt is instantiated by an order on variables.\<close>
fun lpo_closure :: "('f, 'v) term rel \<Rightarrow> ('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<times> nat \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool" where
  "lpo_closure gt prec s t = (
    if (s, t) \<in> gt then (True, True)
    else (
      case (s,t) of
        (Var x, Var y) \<Rightarrow> (False, x = y)
      | (Var x, Fun g ts) \<Rightarrow> (False, ts = [] \<and> (snd prec) (g,0))
      | (Fun f ss, Var y) \<Rightarrow> (let con = (\<exists> s \<in> set ss. snd (lpo_closure gt prec s (Var y))) in (con,con))
      | (Fun f ss, Fun g ts) \<Rightarrow>
         if (\<exists> s \<in> set ss. snd (lpo_closure gt prec s (Fun g ts)))
         then (True,True)
         else (case (fst prec) (f,length ss) (g,length ts) of (prs,prns) \<Rightarrow>
           if prns \<and> (\<forall> t \<in> set ts. fst (lpo_closure gt prec (Fun f ss) t))
           then if prs
              then (True,True) 
              else lex_ext_unbounded (lpo_closure gt prec) ss ts
           else (False,False)))
    )"

abbreviation S where "S p cs s t \<equiv> fst (rpo_unbounded p cs s t)"

abbreviation NS where "NS p cs s t \<equiv> snd (rpo_unbounded p cs s t)"


lemma lex_ext_unbounded_map2:
  assumes S: "\<And> i. i < length ss \<Longrightarrow> i < length ts \<Longrightarrow> fst (r (ss ! i) (ts ! i)) \<Longrightarrow> fst (q (map f ss ! i) (map f ts ! i))"
  and NS: "\<And> i. i < length ss \<Longrightarrow> i < length ts \<Longrightarrow> snd (r (ss ! i) (ts ! i)) \<Longrightarrow> snd (q (map f ss ! i) (map f ts ! i))"
  shows "(fst (lex_ext_unbounded r ss ts) \<longrightarrow> fst (lex_ext_unbounded q (map f ss) (map f ts)))
    \<and> (snd (lex_ext_unbounded r ss ts) \<longrightarrow> snd (lex_ext_unbounded q (map f ss) (map f ts)))"
  using S NS unfolding lex_ext_unbounded_iff by auto

lemma lpo_lpo_closure:
  fixes prec :: "('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<times> nat \<Rightarrow> bool)"
  assumes lpo: "cs = (\<lambda>_. Lex)"
  shows "((S prec cs (s :: ('f, 'v) term) t) \<longrightarrow> fst (lpo_closure ord prec s t)) \<and>
    (NS prec cs s t \<longrightarrow> snd (lpo_closure ord prec s t))"
  using assms
proof (induct s t rule: rpo_unbounded.induct)
  case (1 pr c x y)
  then show ?case by simp
next
  case (2 pr c x g ts)
  then show ?case by simp
next
  case (3 prec cs f ss y)
  { assume "S prec cs (Fun f ss) (Var y) \<or> NS prec cs (Fun f ss) (Var y)"
    from this[simplified] obtain s where s:"s \<in> set ss" "NS prec cs s (Var y)"
      unfolding Let_def by force
    from 3(1)[OF s(1) 3(2)] s have s:"\<exists>s \<in> set ss. snd (lpo_closure ord prec s (Var y))" by blast
    let ?lc = "lpo_closure ord prec (Fun f ss) (Var y)"
    have "fst ?lc \<and> snd ?lc" proof(cases "(Fun f ss, Var y) \<in> ord")
      case False
      hence no_ord:"(Fun f ss, Var y) \<in> ord = False" by auto
      from s False show ?thesis
        unfolding lpo_closure.simps[of _ _ _ "Var y"] no_ord if_False by auto
    qed (auto)
  }
  thus ?case by fast
next
  case (4 prec cs f ss g ts)
  let ?lc = "lpo_closure ord prec (Fun f ss) (Fun g ts)"
  note lc_simps = lpo_closure.simps[of _ _ "Fun f ss" "Fun g ts", of ord prec]
  { assume S:"S prec cs (Fun f ss) (Fun g ts)"
    have "fst ?lc" proof(cases "(Fun f ss, Fun g ts) \<in> ord")
      case False
      hence no_ord:"(Fun f ss, Fun g ts) \<in> ord = False" by auto
      note S = S[simplified]
      show ?thesis proof(cases "\<exists>s \<in> set ss. snd (lpo_closure ord prec s (Fun g ts))")
        case True
        thus ?thesis unfolding lc_simps no_ord if_False Let_def by simp
      next
        case False
        note no_subtc = this
        with 4(1)[OF _ 4(5)] have no_subt:"\<not> (\<exists>s \<in> set ss. NS prec cs s (Fun g ts))" by simp
        note S = S[unfolded False[unfolded eq_False[symmetric]] if_False]
        note lc_simps = lc_simps[unfolded no_ord if_False split_beta fst_conv snd_conv term.case]
        let ?prec = "fst prec (f, length ss) (g, length ts)"
        obtain pr prns where prec[simp]:"fst ?prec = pr" "snd ?prec = prns"
          by (cases "(fst prec (f, length ss) (g, length ts))", simp)
        show ?thesis proof(cases prns)
          case True
          note weak_prec = this
          show ?thesis proof(cases "\<forall>t\<in>set ts. S prec cs (Fun f ss) t")
            case True
            with 4(2)[OF no_subt prod.collapse _ 4(5)]
            have all_ti_smaller:"\<forall>t\<in>set ts. fst (lpo_closure ord prec (Fun f ss) t)"
              unfolding prec[symmetric] by simp
            show ?thesis proof(cases pr)
              case True
              from True False weak_prec all_ti_smaller show ?thesis unfolding lc_simps prec by simp
            next
              case False
              with lc_simps True no_subtc weak_prec all_ti_smaller
              have lc:"?lc = lex_ext_unbounded (lpo_closure ord prec) ss ts" by simp
              from prec have prec':"fst prec (f, length ss) (g, length ts) = (pr, prns)"
                by (smt prod.collapse)
              from True False S no_subt have S:"fst (lex_ext_unbounded (rpo_unbounded prec cs) ss ts)"
                unfolding prec' prod.case 4(5) by (simp add: weak_prec)
              from weak_prec True have "prns \<and> (\<forall>t\<in>set ts. S prec cs (Fun f ss) t)" by simp
              note IH = 4(4)[OF no_subt prod.collapse, unfolded prec, OF this False, unfolded 4(5)]
              note IH = IH[OF refl order_tag.distinct(1) _ _ refl]
              note lex_ext_mono = lex_ext_unbounded_mono[of ss ts "rpo_unbounded prec cs"]
              note lex_ext_mono = lex_ext_mono[of "lpo_closure ord prec", unfolded 4(5)]
              from this IH S show ?thesis unfolding 4(5) unfolding lc by simp
            qed
        next
          case False
          with S no_subt lpo show ?thesis by auto
        qed
      next
        case False
          with S no_subt lpo show ?thesis by (simp add: prod.case_eq_if)
        qed
      qed
    qed (auto)
  } note S = this
  { assume NS:"NS prec cs (Fun f ss) (Fun g ts)"
    have "snd ?lc" proof(cases "(Fun f ss, Fun g ts) \<in> ord")
      case False
      hence no_ord:"(Fun f ss, Fun g ts) \<in> ord = False" by auto
      note NS = NS[simplified]
      show ?thesis proof(cases "\<exists>s \<in> set ss. snd (lpo_closure ord prec s (Fun g ts))")
        case True
        thus ?thesis unfolding lc_simps no_ord if_False Let_def by simp
      next
        case False
        note no_subtc = this
        with 4(1)[OF _ 4(5)] have no_subt:"\<not> (\<exists>s \<in> set ss. NS prec cs s (Fun g ts))" by simp
        note NS = NS[unfolded False[unfolded eq_False[symmetric]] if_False]
        note lc_simps = lc_simps[unfolded no_ord if_False split_beta fst_conv snd_conv term.case]
        let ?prec = "fst prec (f, length ss) (g, length ts)"
        obtain pr prns where prec[simp]:"fst ?prec = pr" "snd ?prec = prns"
          by (cases "(fst prec (f, length ss) (g, length ts))", simp)
        show ?thesis proof(cases prns)
          case True
          note weak_prec = this
          show ?thesis proof(cases "\<forall>t\<in>set ts. S prec cs (Fun f ss) t")
            case True
            with 4(2)[OF no_subt prod.collapse _ 4(5)]
            have all_ti_smaller:"\<forall>t\<in>set ts. fst (lpo_closure ord prec (Fun f ss) t)"
              unfolding prec[symmetric] by simp
            show ?thesis proof(cases pr)
              case True
              from True False weak_prec all_ti_smaller show ?thesis unfolding lc_simps prec by simp
            next
              case False
              with lc_simps True no_subtc weak_prec all_ti_smaller
              have lc:"?lc = lex_ext_unbounded (lpo_closure ord prec) ss ts" by simp
              from prec have prec':"fst prec (f, length ss) (g, length ts) = (pr, prns)"
                by (smt prod.collapse)
              from True False NS no_subt have NS:"snd (lex_ext_unbounded (rpo_unbounded prec cs) ss ts)"
                unfolding prec' prod.case 4(5) by (simp add: weak_prec)
              from weak_prec True have "prns \<and> (\<forall>t\<in>set ts. S prec cs (Fun f ss) t)" by simp
              note IH = 4(4)[OF no_subt prod.collapse, unfolded prec, OF this False, unfolded 4(5)]
              note IH = IH[OF refl order_tag.distinct(1) _ _ refl]
              note lex_ext_mono = lex_ext_unbounded_mono[of ss ts "rpo_unbounded prec cs"]
              note lex_ext_mono = lex_ext_mono[of "lpo_closure ord prec", unfolded 4(5)]
              from this IH NS show ?thesis unfolding 4(5) unfolding lc by simp
            qed
        next
          case False
          with NS no_subt lpo show ?thesis by auto
        qed
      next
        case False
          with NS no_subt lpo show ?thesis by (simp add: prod.case_eq_if)
        qed
      qed
    qed (auto)
  }
  with S show ?case by simp
qed

end
end
end
